@@ -3,7 +3,7 @@
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
     2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-    2017, 2018, 2019 Massachusetts Institute of Technology
+    2017, 2018, 2019, 2020 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -26,7 +26,7 @@ USA.
 
 ;;;; 6.001: HP-UX Floppy Commands
 
-(declare (usual-integrations))
+;(declare (usual-integrations))
 
 (define (run-floppy-login-loop)
   (set! floppy-contents-loaded? false)
@@ -449,7 +449,7 @@ then answer \"yes\" to the prompt below.")
   (append-string "Copying files from working directory to floppy.\n")
   (let* ((working-directory (read-working-directory))
 	 (floppy-directory (read-floppy-directory)))
-    (with-values
+    (call-with-values
 	(lambda ()
 	  (three-way-sort file-record/name=?
 			  working-directory
@@ -620,7 +620,7 @@ M-x rename-file, or use the `r' command in Dired.")
 	      (let ((eol
 		     (or (string-find-next-char string #\newline start end)
 			 end)))
-		(with-values
+		(call-with-values
 		    (lambda ()
 		      (parse-dosls-line string start eol offset))
 		  (lambda (filename time)
@@ -867,13 +867,13 @@ M-x rename-file, or use the `r' command in Dired.")
 	  (values '() '() set*)
 	  (let ((item (member? (car set) set*)))
 	    (if item
-		(with-values
+		(call-with-values
 		    (lambda () (loop (cdr set) (delq! (car item) set*)))
 		  (lambda (set-only both set*-only)
 		    (values set-only
 			    (cons (cons (car set) (car item)) both)
 			    set*-only)))
-		(with-values (lambda () (loop (cdr set) set*))
+		(call-with-values (lambda () (loop (cdr set) set*))
 		  (lambda (set-only both set*-only)
 		    (values (cons (car set) set-only)
 			    both
