@@ -3,7 +3,7 @@
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
     2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-    2017, 2018, 2019 Massachusetts Institute of Technology
+    2017, 2018, 2019, 2020 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -747,31 +747,27 @@ USA.
   ;; lower case or capitalized.  This is a very general definition of
   ;; capitalization; if you need something more specific you should
   ;; call this procedure on the individual words.
-  (letrec
-      ((find-first-word
-	(lambda (start)
-	  (and (fix:< start end)
-	       (let ((char (string-ref string start)))
-		 (if (char-upper-case? char)
-		     (scan-word-tail (fix:+ start 1))
-		     (and (not (char-lower-case? char))
-			  (find-first-word (fix:+ start 1))))))))
-       (scan-word-tail
-	(lambda (start)
-	  (or (fix:= start end)
-	      (let ((char (string-ref string start)))
-		(if (char-lower-case? char)
-		    (scan-word-tail (fix:+ start 1))
-		    (and (not (char-upper-case? char))
-			 (find-subsequent-word (fix:+ start 1))))))))
-       (find-subsequent-word
-	(lambda (start)
-	  (or (fix:= start end)
-	      (let ((char (string-ref string start)))
-		(if (char-alphabetic? char)
-		    (scan-word-tail (fix:+ start 1))
-		    (find-subsequent-word (fix:+ start 1))))))))
-    (find-first-word start)))
+  (define (find-first-word start)
+    (and (fix:< start end)
+	 (let ((char (string-ref string start)))
+	   (if (char-upper-case? char)
+	       (scan-word-tail (fix:+ start 1))
+	       (and (not (char-lower-case? char))
+		    (find-first-word (fix:+ start 1)))))))
+  (define (scan-word-tail start)
+    (or (fix:= start end)
+	(let ((char (string-ref string start)))
+	  (if (char-lower-case? char)
+	      (scan-word-tail (fix:+ start 1))
+	      (and (not (char-upper-case? char))
+		   (find-subsequent-word (fix:+ start 1)))))))
+  (define (find-subsequent-word start)
+    (or (fix:= start end)
+	(let ((char (string-ref string start)))
+	  (if (char-alphabetic? char)
+	      (scan-word-tail (fix:+ start 1))
+	      (find-subsequent-word (fix:+ start 1))))))
+  (find-first-word start))
 
 (define (string-capitalize string)
   (guarantee-string string 'STRING-CAPITALIZE)
