@@ -3,7 +3,7 @@
 Copyright (C) 1986, 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994,
     1995, 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
     2006, 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
-    2017, 2018, 2019 Massachusetts Institute of Technology
+    2017, 2018, 2019, 2020 Massachusetts Institute of Technology
 
 This file is part of MIT/GNU Scheme.
 
@@ -2236,7 +2236,7 @@ WARNING: With a prefix argument, this command may take a very long
 
 (define ((probe-folder-output-processor folder))
   (let ((folder (weak-car folder)))
-    (and folder
+    (and (folder? folder)
 	 (if (and (imail-folder->buffer folder #f)
 		  (eq? (folder-connection-status folder) 'ONLINE))
 	     (begin
@@ -2307,14 +2307,16 @@ WARNING: With a prefix argument, this command may take a very long
 	   (cond ((ref-variable imail-kept-headers context)
 		  => (lambda (regexps)
 		       (remove-duplicates!
-			(append-map*!
+			(fold-right
+			 (lambda (regexp result)
+			   (append! (filter (lambda (header)
+					      (re-string-match
+					       regexp
+					       (header-field-name header)
+					       #t))
+					    headers)
+				    result))
 			 (mime-headers)
-			 (lambda (regexp)
-			   (filter (lambda (header)
-				     (re-string-match regexp
-						      (header-field-name header)
-						      #t))
-				   headers))
 			 regexps)
 			(lambda (a b) (eq? a b)))))
 		 ((ref-variable imail-ignored-headers context)
